/*
 *      ISIS release V1.1, Dec. 1988
 *      Export restrictions apply
 */
/*
 *      TCP code for packet transmission from ISIS to client
 *      Performance considerations lead to a somewhat complicated
 *      protocol.
 */

# include "pr.h"
#ifdef UNIX_DOM
# include <sys/un.h>
#endif
# include <fcntl.h>

struct  hostent *gethostbyname();
struct  servent *getservbyname();

static  connect_socket;
static  CLIENT_PORT = 1250;               /* Overrides default if non-zero */

typedef struct cldesc
{
        short   cl_flag;        /* Set if the client is unable to accept data */
        short   cl_pid;         /* Process id of this client */
        address cl_paddr;       /* Address of this client */
        int     cl_queuelen;    /* Number of qnoded bytes when congested */
        qnode   *cl_msgqueue;   /* The congestion qnode */
        qnode   *cl_pmonitor;   /* Remote processes this client is monitoring */
        message *cl_curmsg;     /* Message being transmitted now */
        message *cl_mp;         /* Message most recently received */
        block_desc *cl_blkdesc; /* Message being received now, if CLIENT_READING set */
        int     cl_rlen;        /* Remaining length of message being read */
        char    *cl_rptr;       /* Pointer into message read area */
        iovec   *cl_curiov;     /* copy of the iovec */
        iovec   *cl_iovptr;     /* ptr to current iovec entry */
        int     cl_iovlen;      /* Remaining iovec len */
        int     cl_nbytes;      /* Remaining bytes to send */
        int     cl_lastact;     /* Time when last active */
        bitvec  cl_watchedby;   /* Sites watching this client */
} cldesc;

#define CLIENT_BUSY         0x01    /* Client is busy */
#define CLIENT_READING      0x02    /* Reading a message from this client */
#define CLIENT_PMONITOR     0x04    /* Client is monitoring one or more remote processes */
#define CLIENT_MALLOC       0x08    /* Client is monitoring one or more remote processes */

static cldesc   cl[MAXBITS];

client_init()
  {
        static saddr my_addr;
#ifdef  UNIX_DOM
        struct sockaddr_un ux_addr;
#endif

        /* Create a TCP socket to listen on */
#ifdef  UNIX_DOM
        if((connect_socket = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
        {
            perror("socket");
            panic("client_init");
        }
        connect_bit = connect_socket;
        bis(input_mask, connect_socket);
        if(CLIENT_PORT == 0)
        {
            register struct servent *sp = getservbyname("isis", "tcp");
            if(sp == (struct servent*)0)
                panic("isis.*: service not listed in /etc/services on this host");
            CLIENT_PORT = sp->s_port;
        }
        ux_addr.sun_family = AF_UNIX;
        if(connect_port[my_site_no])
            sprintf(ux_addr.sun_path, "/tmp/Is%d", connect_port[my_site_no]);
        else
            sprintf(ux_addr.sun_path, "/tmp/Is%d", CLIENT_PORT);
        unlink(ux_addr.sun_path);
        if(bind(connect_socket, (struct sockaddr*)&ux_addr, sizeof(my_addr)) == -1)
        {
            perror("bind");
            panic("client_init unable to create <%s>.  Is ISIS already running?\n", ux_addr.sun_path);
        }
#else   UNIX_DOM
        if((connect_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1)
        {
            perror("socket");
            panic("client_init");
        }
        connect_bit = connect_socket;
        bis(input_mask, connect_socket);
        if(CLIENT_PORT == 0)
        {
            register struct servent *sp = getservbyname("isis", "tcp");
            if(sp == (struct servent*)0)
                panic("isis.*: service not listed in /etc/services on this host");
            CLIENT_PORT = sp->s_port;
        }
        my_addr.sin_family = AF_INET;
        if(connect_port[my_site_no])
            my_addr.sin_port = htons(connect_port[my_site_no]);
        else
            my_addr.sin_port = htons(CLIENT_PORT);
        if(bind(connect_socket, (struct sockaddr*)&my_addr, sizeof(my_addr)) == -1)
        {
            perror("bind connect_socket");
            panic("client_init unable to acquire port %d.  Is ISIS already running?\n", ntohs(my_addr.sin_port));
        }
#endif  UNIX_DOM
        listen(connect_socket, 5);
  }

/* Select shows that we should be able to do an accept now */
client_accept()
  {
        register snum = accept(connect_socket, (struct sockaddr *)0, (int*)0);
        register message *msg;
        cinfo ci;

        if(snum == -1)
        {
            perror("accept");
            panic("accept failed! max_cl %d", max_cl);
        }
        bis(input_mask, snum);
        if(snum > max_cl)
            max_cl = snum;
        if(cl[snum].cl_pid)
            client_failed(snum, FALSE);
#       if (SUN)
            setsockopt(snum, SOL_SOCKET, SO_DONTLINGER, (char*)0, (int*)0);
#       endif
        if(snum >= MAXBITS)
            panic("socket_no > MAXBITS");
        fcntl(snum, F_SETFL, FNDELAY);
        ci.ci_my_site_no = my_site_no;
        ci.ci_my_site_incarn = my_site_incarn;
        ci.ci_my_site_id = my_site_id;
        ci.ci_cur_sview = current_view;
        if((ci.ci_genid = snum) == 0 || snum > 127)
            panic("Unexpected fdes %d in client_accept", snum);
        strcpy(ci.ci_isisdir, isis_dir);
        bcopy(site_names, ci.ci_site_names, sizeof(site_names));
        msg = msg_genmsg(FLD_ANSW, (char*)&ci, FTYPE_CHAR, sizeof(ci), 0);
        if(tcp_write(snum, msg, 1) == -1)
        {
            close(snum);
            bic(input_mask, snum);
        }
        msg_delete(msg);
  }

/* Read a message from a local file descriptor */
pr_local_rcv(sd)
 {
        register cldesc *cp = &cl[sd];
        register message *mp;
        register long *space;
        long length = 0;
        register nb, eof = 1;

        if((cp->cl_flag&CLIENT_READING) == 0)
        {
            cp->cl_rlen = 0;
            if((nb = read(sd, (char*)&cp->cl_rlen, sizeof(long))) != sizeof(long))
                goto bad;
            eof = 0;
            cp->cl_rlen = ntohl (cp->cl_rlen);
            cp->cl_blkdesc = msg_blockalloc (cp->cl_rlen, 0);
            cp->cl_flag |= CLIENT_READING;
            space = (long *) cp->cl_blkdesc->addr;
            space[0] = htonl (cp->cl_rlen);
            cp->cl_rptr = (char*)&space[1];
            cp->cl_rlen -= sizeof(long);
        }
        else
            space = (long *) cp->cl_blkdesc->addr;
        while(cp->cl_rlen)
        {
            if((nb = read(sd, cp->cl_rptr, cp->cl_rlen)) <= 0)
            {
                if(nb == 0 && eof)
                    goto bad;
                if(nb == 0 || (nb == -1 && errno == EWOULDBLOCK))
                    return;
                goto bad;
            }
            eof = 0;
            cp->cl_rlen -= nb;
            cp->cl_rptr += nb;
        }
        if(cp->cl_pid && cp->cl_pid != ISIS && cp->cl_paddr.site == 0)
            /* He was too quick */
            goto bad;
        if(mp = msg_reconstruct((char*)space, cp->cl_blkdesc))
        {
            EVENT(S_SYSCALLS);
            if(cp->cl_pid > 0)
            {
                address sender, msender;
                sender = cp->cl_paddr;
                msender = msg_getsender(mp);
                sender.entry = msender.entry;
                msg_setsender(mp, sender);
            }
            if(msg_tracemsgs)
            {
                int pno = PN_CLIENT;
                msg_replacefield(mp, SYSFLD_PROTO, (char*)&pno, FTYPE_LONG, sizeof(int));
            }
            cp->cl_flag &= ~CLIENT_READING;
            cp->cl_blkdesc = 0;
            cp->cl_rlen = 0;
            cp->cl_rptr = 0;
            begin
            {
                register cldesc *ccp;
                /* Inefficient, but all I could come up with! */
                for(ccp = cl; ccp <= &cl[max_cl]; ccp++)
                    if(ccp->cl_mp == mp)
                        ccp->cl_mp = 0;
                cp->cl_mp = mp;
            }
            pr_local_delivery(mp, (int*)0);
            msg_delete(mp);
            return;
        }
  bad:
        client_failed(sd, TRUE);
  }

/* Write a message to a local file descriptor, blocks until done */
pr_local_send(pid, mp)
  register pid;
  register message *mp;
  {
        register sd, incflag = 1;
        register address *ap;
        address pname;

        sd = client_lookup(pid);
        pname = ADDRESS(my_site_no, my_site_incarn, pid, 0);
        if(pg_verify(pname, mp) == -1)
            return;
        for(ap = msg_getdests(mp); !addr_isnull(*ap); ap++)
            if(addr_cmp(*ap, pname) == 0)
                switch(ap->entry)
                {
                  message *gmgr_spy();
                  case GENERIC_ADDMEMB:
                  case GENERIC_ADDCLIENT:
                  case GENERIC_DELETE:
                  case GENERIC_SIGNAL:
                    if(mp = gmgr_spy(*ap, mp, sd))
                    {
                        incflag = 0;
                        goto deliver;
                    }
                    return;

                  case GENERIC_IGNORE:
                    reply(mp, (char*)0, 0, 0);
                    return;
                }
  deliver:
        if(sd == -1)
        {
            register n;
            address pname;
            if(msg_getfield(mp, SYSFLD_MSGID, 1, (int*)0))
            {
                pname = ADDRESS(my_site_no, my_site_incarn, pid, 0);
                reply_as(&pname, mp, (char*)0, 0, 0);
            }
        }
        else
        {
            if(pb_checkempty(pid))
                 msg_insertfield(mp, SYSFLD_BYPASS, (char*)0, 0, 0);
            tcp_write(sd, mp, incflag);
        }
  }

/* Informs client that it can resume using bypass broadcasts */
pb_isempty (pid)
  {
        address paddr;
	register sd;
        register message *mp;
#ifdef	BYPASS
        if((sd = client_lookup(pid)) == -1)
            return;
        mp = msg_newmsg();
        paddr = ADDRESS(my_site_no, my_site_incarn, pid, GENERIC_BYCNTL);
        msg_setdest(mp, paddr);
        tcp_write(sd, mp, 1);
        msg_delete(mp);
#endif
  }

#define LONG_IOV        64              /* Longer than this -> use malloc() */

adesc iov_ad ={ sizeof(iovec)*LONG_IOV, 0, 4 };

#define iov_alloc()     ((iovec*)mallocate(&iov_ad))
#define iov_free(iovp)  mdeallocate((char*)iovp, &iov_ad)

#ifdef  SIMWRITEV
writev(so, iovp, iovl)
  register iovec *iovp;
  register iovl;
  {
        register nb, wb;
        while(iovl--)
        {
            wb = write(so, iovp->iov_ptr, iovp->iov_len);
            if(wb == -1)
                return(-1);
            nb += wb;
            if(wb < (iovp++)->iov_len)
                break;
        }
        return(nb);
  }
#endif

/* Write to client after spying on the message traffic */
tcp_write(sd, mp, incflag)
  register message *mp;
  {
        register cldesc *cp = &cl[sd];
        register nb, iovlen;
        address sender;

        EVENT(S_CLSENT);
        if(incflag)
            msg_increfcount(mp);
        if(cp->cl_flag&CLIENT_BUSY)
        {
            /* If client is not ready, don't try to send yet */
            client_enqueue(sd, mp);
            return(0);
        }
        sender = msg_getreplyto(mp);
        bis(cp->cl_watchedby, sender.site);
        cp->cl_curmsg = mp;
        cp->cl_nbytes = msg_getlen(mp);
        cp->cl_iovlen = msg_getiovlen(mp);
        nb = sizeof(iovec)*cp->cl_iovlen;
        if(cp->cl_iovlen > LONG_IOV)
        {
            cp->cl_curiov = (iovec*)malloc(nb = sizeof(iovec)*cp->cl_iovlen);
            cp->cl_flag |= CLIENT_MALLOC;
        }
        else
            cp->cl_curiov = iov_alloc();
        bcopy((char*)msg_getiovec(mp), (char*)cp->cl_curiov, nb);
        cp->cl_iovptr = cp->cl_curiov;
        /* Respect the system-wide limit on iovec lengths */
        if((iovlen = cp->cl_iovlen) > MAX_IOVLEN)
             iovlen = MAX_IOVLEN;
        nb = writev(sd, cp->cl_iovptr, iovlen);
        if(nb == -1 && errno != EWOULDBLOCK)
        {
            print("tcp_write failed to client %d\n", cp->cl_pid);
            client_failed(sd, TRUE);
            return(-1);
        }
	cl_xfered(cp, nb);
        return(0);
  }

cl_restart_io(sdlist)
  bitvec sdlist;
  {
        register sd, nb;

        bicv(congested, sdlist);
        for(sd = 0; sd <= max_cl; sd++)
            if(bit(sdlist, sd))
            {
                register cldesc *cp = &cl[sd];
                nb = writev(sd, cp->cl_iovptr, cp->cl_iovlen);
                if(nb == -1 && errno != EWOULDBLOCK)
                    client_failed(sd, TRUE);
                else
                {
                    cl_xfered(cp, nb);
                    while(cp->cl_pid && (cp->cl_flag&CLIENT_BUSY) == 0 && cp->cl_queuelen)
                    {
                        register qnode *qp = qu_head(cp->cl_msgqueue);
                        register message *mp;
                        mp = qp->qu_msg;
                        cp->cl_queuelen -= msg_getlen(mp);
                        tcp_write(sd, mp, 1);
                        /* Free qnode node and message, both at once */
                        qu_free(qp);
                    }
                }
            }
  }

cl_xfered(cp, nb)
  register cldesc *cp;
  register nb;
  {
        /* Better safe than sorry */
        t_scheck();
        cp->cl_lastact = Gettime();
        if(cp->cl_curmsg == 0)
            return;
        if(nb == -1)
            nb = 0;
        if((cp->cl_nbytes -= nb) == 0)
        {
            if(cp->cl_flag&CLIENT_MALLOC)
                free(cp->cl_curiov);
            else
                iov_free(cp->cl_curiov);
            msg_delete(cp->cl_curmsg);
            cp->cl_flag &= ~CLIENT_MALLOC;
            cp->cl_curiov = 0;
            cp->cl_curmsg = 0;
            cp->cl_flag &= ~CLIENT_BUSY;
            return;
        }
        bis(congested, cp-cl);
        cp->cl_flag |= CLIENT_BUSY;
        while(nb > cp->cl_iovptr->iov_len)
        {
            nb -= cp->cl_iovptr->iov_len;
            --cp->cl_iovlen;
            ++cp->cl_iovptr;
        }
        if(nb)
        {
            cp->cl_iovptr->iov_base += nb;
            cp->cl_iovptr->iov_len -= nb;
        }
  }

condition want_incarn;
static  cl_congested;

/* Registers PID with system */
cl_register(mp)
  register message *mp;
  {
        register *pid = (int*)msg_getfield(mp, CL_PID, 1, (int*)0);
        register cldesc *cp;
        for(cp = cl; cp <= &cl[max_cl]; cp++)
            if(cp->cl_pid == pid[0])
            {
                print("Attempt to re-register process %d\n", pid[0]);
                break;
            }
        switch(pid[0])
        {
          default:
            if(pid[0] < 0)
            {
                print("Attempt to register as pid %d -- rejected\n", pid[0]);
                cp = 0;
            }
            break;

          case ISIS:
          case REXEC:
          case RMGR:
          case NEWS:
          case XMGR:
            break;
        }
        if(cp > &cl[max_cl])
        {
            for(cp = cl; cp <= &cl[max_cl]; cp++)
                if(cp->cl_mp == mp)
                    break;
            cp->cl_pid = pid[0];
            bclr(cp->cl_watchedby);
            if(cp->cl_msgqueue == 0)
                cp->cl_msgqueue = qu_null();
            cp->cl_pmonitor = qu_null();
            cp->cl_paddr = NULLADDRESS;
        }
        if(pid[0] != ISIS && my_site_incarn == RECOVERY_INCARN)
            t_wait(&want_incarn, "want_incarn");
        cp->cl_paddr = ADDRESS(my_site_no, my_site_incarn, pid[0], 0);
        cp->cl_paddr.portno = pid[1];
        reply(mp, (char*)&my_site_incarn, FTYPE_LONG, sizeof(int));
        if(cl_congested)
        {
            register message *cmsg = msg_newmsg();
            address addr;
            addr = cp->cl_paddr;
            addr.entry = GENERIC_CONGESTED;
            msg_setdest(cmsg, addr);
            pr_local_send(pid, cmsg);
        }
  }

cl_send_all(ent, msg)
  register ent;
  register message *msg;
  {
        register cldesc *cp;
        address addr;
        register message *mp = msg;
        if(mp == (message*)0)
            mp = msg_newmsg();
        cl_congested = (ent == GENERIC_CONGESTED);
        for(cp = cl; cp <= &cl[max_cl]; cp++)
             if(cp->cl_pid)
             {
                addr = cp->cl_paddr;
                addr.entry = ent;
                msg_setdest(mp, addr);
                pr_local_send(cp->cl_pid, mp);
             }
        if(msg == (message*)0)
            msg_delete(mp);
  }

cl_pmonitor(mp)
  register message *mp;
  {
        register address *ap = (address*)msg_getfield(mp, CL_PNAME, 1, (int*)0);
        register cldesc *cp;
        int outcome;
        for(cp = cl; cp <= &cl[max_cl]; cp++)
            if(cp->cl_mp == mp)
                break;
        cp->cl_mp = 0;
        ap->entry = 0;
        cp->cl_flag |= CLIENT_PMONITOR;
        if(is_proc_alive(*ap) && cp->cl_pmonitor)
        {
            outcome = 0;
            if(pg_find(cp->cl_pmonitor, *ap) == (qnode*)0)
                pg_add(cp->cl_pmonitor, *ap, 0, nullroutine);
        }
        else
            outcome = -1;
        reply(mp, &outcome, FTYPE_LONG, sizeof(outcome));
  }

/* Called when first adding a proc node: actually check with remote site */
is_proc_alive(addr)
  address addr;
  {
        if(addr.site == my_site_no)
        {
            register cldesc *cp;
            if(addr.incarn != my_site_incarn)
                return(0);
            if(check_proc_status(addr.process, 0))
                return(1);
            tsleep(2);
            if(check_proc_status(addr.process, 0))
                return(1);
            tsleep(20);
            return(check_proc_status(addr.process, 0));
        }
        else if(current_view.sv_incarn[addr.site] != addr.incarn)
            return(0);
        begin
        {
            register message *mp;
            char answ;
            static address dest[2];
            mp = msg_genmsg(CL_PNAME, &addr, FTYPE_ADDRESS, sizeof(address), 0);
            dest[0] = addr;
            dest[0].process = PROTOCOLS;
            dest[0].entry = PR_IS_ALIVE;
            if(BCAST(dest, mp, 1, collect_answ, &answ, 1) == 0 || answ == '-')
            {
                msg_delete(mp);
                return(0);
            }
            msg_delete(mp);
        }
        return(1);
  }

pr_is_alive(mp)
  register message *mp;
  {
        register address *ap = (address*)msg_getfield(mp, CL_PNAME, 1, (int*)0);
        address msender;
        msender = msg_getsender(mp);
        if(ap->site == my_site_no && ap->incarn == my_site_incarn)
        {
            if(check_proc_status(ap->process, msender.site))
                goto ok;
            tsleep(2); 
            if(check_proc_status(ap->process, msender.site))
                goto ok;
            /* Currently unknown, but give him a chance to start just in case... */ 
            tsleep(20); 
            if(check_proc_status(ap->process, msender.site))
            {
            ok:
                reply(mp, "+", FTYPE_CHAR, 1);
                return;
            }
        } 
        reply(mp, "-", FTYPE_CHAR, 1);
  }

check_proc_status(pid, s)
  register pid;
  {
        register cldesc *cp;
        if(pid == PROTOCOLS)
            return(1);
        for(cp = cl; cp <= &cl[max_cl]; cp++)
            if(cp->cl_pid == pid)
            {
                if(s)
                    bis(cp->cl_watchedby, s);
                return(1);
            }
        return(0);
  }

/* Called when a process has been observed to fail */
proc_has_failed(pname)
  address pname;
  {
        register cldesc *cp;
        register qnode *qp;
        for(cp = cl; cp <= &cl[max_cl]; cp++)
            if((cp->cl_flag&CLIENT_PMONITOR) && (qp = pg_find(cp->cl_pmonitor, pname)))
            {
                static address addrs[2];
                register message *mp = msg_genmsg(CL_PNAME, &pname, FTYPE_ADDRESS, sizeof(address), 0);
                qu_free(qp);
                addrs[0] = cp->cl_paddr;
                addrs[0].entry = GENERIC_PROC_FAILED;
                msg_setdests(mp, addrs);
                pr_bcast(mp);
                msg_delete(mp);
                if(qu_head(cp->cl_pmonitor) == 0)
                    cp->cl_flag &= ~CLIENT_PMONITOR;
            }
  }

client_lookup(pid)
  register pid;
  {
        register sd;
        for(sd = 0; sd <= max_cl; sd++)
            if(cl[sd].cl_pid == pid)
                return(sd);
        return(-1);
  }

client_newview()
  {
        register cldesc *cp;
        register qnode *qp, *nqp;
        for(cp = cl; cp <= &cl[max_cl]; cp++)
        {
            address addr;
            register message *mp = msg_genmsg(CL_VIEW, &current_view, FTYPE_CHAR, sizeof(sview), 0);
            addr = cp->cl_paddr;
            addr.entry = GENERIC_NEW_SVIEW;
            msg_setdest(mp, addr);
            pr_bcast(mp);
            msg_delete(mp);
            if(cp->cl_flag&CLIENT_PMONITOR)
                for(qp = cp->cl_pmonitor->qu_next; qp != cp->cl_pmonitor; qp = nqp)
                    if(current_view.sv_incarn[qp->qu_pname.site] != qp->qu_pname.incarn)
                    {
                        /* New view will trigger this watch event in the client */
                        nqp = qp->qu_next;
                        qu_free(qp);
                        if(qu_head(cp->cl_pmonitor) == 0)
                            cp->cl_flag &= ~CLIENT_PMONITOR;
                    }
                    else
                        nqp = qp->qu_next;
        }
  }


typedef struct
{
        int     fi_pid;
        bitvec  fi_watchedby;
} failinfo;

adesc fi_ad ={ sizeof(failinfo), 0, 4 };

#define fi_alloc()   ((failinfo*)mallocate(&fi_ad))
#define fi_free(fip)  mdeallocate((char*)fip, &fi_ad)

client_failed(sd, closef)
  register sd;
  {
        int pid, client_crashed();
        register cldesc *cp = &cl[sd];
        register failinfo *fip;
        char clfile[32];

        if(closef)
        {
            bic(input_mask, sd);
            close(sd);
        }
        if((pid = cp->cl_pid) == 0)
            return;
        if(pid == ISIS)
            pr_dump(DUMP_ALL, "monitoring process at this site has crashed!");
#ifdef  UNIX_DOM
        sprintf(clfile, "/tmp/Cl%d", pid);
        unlink(clfile);
#endif
        cp->cl_pid = 0;
        if(cp->cl_msgqueue)
            qu_freeall(cp->cl_msgqueue);
        if(cp->cl_pmonitor)
            qu_freeall(cp->cl_pmonitor);
        if(cp->cl_curiov)
        {
            if(cp->cl_flag&CLIENT_MALLOC)
                free(cp->cl_curiov);
            else
                iov_free(cp->cl_curiov);
        }
        if(cp->cl_curmsg)
            msg_delete(cp->cl_curmsg);
        if(cp->cl_blkdesc)
            msg_blockfree (cp->cl_blkdesc);
        fip = fi_alloc();
        fip->fi_pid = pid;
        fip->fi_watchedby = cp->cl_watchedby;
        bzero(cp, sizeof(*cp));
        t_fork(client_crashed, fip, (message*)0);
  }

client_crashed(fip)
  register failinfo *fip;
  {
        register message *mp;
        address pname;
        site_id slist[MAX_SITES];
        register site_id *sp = slist;
        register s, i;
        /* Now inform other sites and initiate failure gbcast */
        cb_procflush(fip->fi_pid);
        for(s = 1; s < MAX_SITES; s++)
            if((s == my_site_no || bit(fip->fi_watchedby, s)) && (i = current_view.sv_incarn[s]) >= 0)
                *sp++ = MAKE_SITE_ID(s, i);
        *sp = 0;
        mp = msg_genmsg(FLD_FAILED, (char*)&fip->fi_pid, FTYPE_LONG, sizeof(int), 0);
        pname = ADDRESS(my_site_no, my_site_incarn, fip->fi_pid, 0);
        msg_setsender(mp, pname);
        BCAST_SL(slist, PROTOCOLS, PR_CLFAILED, mp, 0, 0, 0, 0);
        msg_delete(mp);
        pg_failed(fip->fi_pid);
        fi_free(fip);
  }

client_kill(sd)
  register sd;
  {
        register cldesc *cp = &cl[sd];
        if(cp->cl_pid)
            kill(cp->cl_pid, SIGKILL);
  }

#define CL_ACTIVE       2000            /* Client gets 2 seconds to read something */

client_enqueue(sd, msg)
  register sd;
  register message *msg;
  {
        register cldesc *cp = &cl[sd];
        register len = msg_getlen(msg);

        if((cp->cl_queuelen += len)+cp->cl_nbytes > MAXCQUEUE)
        {
            if(Gettime()-cp->cl_lastact < CL_ACTIVE)
                goto ok;
            print("Message qnode for client %d overflowed: discarding a message\n", cp->cl_pid);
            /* Client will either catch signal or simply barf */
            if(cp->cl_pid)
                kill(cp->cl_pid, SIG_OVERFLOW);
            cp->cl_queuelen -= len;
            reply_as(&cp->cl_paddr, msg, (char*)0, 0, 0);
            msg_delete(msg);
            return;
        }
  ok:
        if(cp->cl_msgqueue == 0)
            cp->cl_msgqueue = qu_null();
        qu_add_mp(cp->cl_msgqueue, QU_MSG, msg, msg_delete);
  }

dump_clients()
  {
        register sd;
        for(sd = 0; sd <= max_cl; sd++)
        {
            register cldesc *cp = &cl[sd];
            if(cp->cl_pid)
            {
                register d = 0;
                register qnode *qp;
                pclient(cp->cl_pid);
                print(": ");
                if(cp->cl_flag&CLIENT_BUSY)
                    print("busy (%d bytes in transit, %d bytes on qnode) ", cp->cl_nbytes, cp->cl_queuelen);
                else
                    print("idle");
                if(cp->cl_flag&CLIENT_READING)
                    print(" reading (need %d bytes)", cp->cl_rlen);
                if(cp->cl_flag&CLIENT_PMONITOR)
                {
                    register qnode *qp;
                    print(" monitoring < ");
                    for(qp = cp->cl_pmonitor->qu_next; qp != cp->cl_pmonitor; qp = qp->qu_next)
                    {
                        paddr(qp->qu_pname);
                        print(" ");
                    }
                    print(">");
                }
                bic(cp->cl_watchedby, my_site_no);
                if(btst(cp->cl_watchedby))
                {
                    register s;
                    print(" watched by sites < ");
                    for(s = 1; s < MAX_SITES; s++)
                        if(bit(cp->cl_watchedby, s) && current_view.sv_incarn[s] != DOWN_INCARN)
                            print("%d ", s);
                    print(">");
                }
                bis(cp->cl_watchedby, my_site_no);
                print("\n");
                if(cp->cl_msgqueue == 0)
                    print("  *** null msgqueue pointer!  how can this be? ***\n");
                else for(qp = cp->cl_msgqueue->qu_next; qp != cp->cl_msgqueue; qp = qp->qu_next)
                {
                    print("  Delivery qnode[%d]: ", ++d);
                    pmsg(qp->qu_msg);
                }
            }
        }
  }

pclient(pid)
  {
        switch(pid)
        {
          default: print("client pid=%d", pid); break;
          case ISIS: print("<site-monitor>"); break;
          case REXEC: print("<rexec>"); break;
          case RMGR: print("<rmgr>"); break;
          case NEWS: print("<news>"); break;
          case XMGR: print("<xmgr>"); break;
        }
  }
